/*
17-5 Lagged Fibonacci random number generator 
with thread independent generators
by Sigismondo Boschi
Agoust 13th, 2009
to be checked! 
Please send feedback if you find any error or realize some improvement
... and either if you check it against some reference and find it correct!
net1701@gmail.com
*/


#include <stdio.h>
#include <math.h>

// includes, kernels
#include <lfrngk.cu>

/* Fortran90 version - neither this is checked however */
/*
parameter(a=16807, r=2836, q=127773, m=z'7fffffff',mask31=z'7fffffff',mask32=z'ffffffff')
subroutine init175(iseed)
integer(kind=4) iseed
integer(kind=4) :: i,hi,lo
do i=1,16
   hi=iseed/q
   lo=mod(iseed,q)
   iseed = a*lo - r*hi
   if (iseed.lt.0) iseed = iseed + m
   ifibreg(i) = ishft(iseed,1)
enddo
ifibreg(11) = ifibreg(11) + 1
ifibreg(17) = 0
lptr = 17
kptr = 5
return
end subroutine

integer function irnd175()
integer(kind=4) :: itemp
itemp = ifibreg(lptr) + ifibreg(kptr)
ifibreg(lptr) = iand(itemp,mask32)
lptr = lptr - 1
kptr = kptr - 1
if (lptr.le.0) lptr = 17
if (kptr.le.0) kptr = 17
irnd175 = iand(ishft(itemp,-1),mask31)
return
end function
*/

unsigned int fibreg[17];
int lptr;
int kptr;
void InitLfg175(int iseed) {
	const int a=16807;
	const int r=2836; 
	const int q=127773;
	for(int i=0; i<16; i++) {
		int hi=iseed/q;
		int lo=iseed % q;
		iseed = a*lo - r*hi;
		if (iseed < 0) iseed += 0x7FFFFFFF;
		fibreg[i] = iseed << 1; 
    }
	fibreg[10] |= 1;
	fibreg[16] = 0;
	lptr=16;
	kptr=4;
}

int Lfg175() {
	unsigned int r = fibreg[lptr] + fibreg[kptr]; 
	fibreg[lptr]=r; 
	lptr = lptr==0 ? 16 : lptr-1;
	kptr = kptr==0 ? 16 : kptr-1;
	return r>>1;
}

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int
main( int argc, char** argv) 
{
	unsigned int local[100];
	int totalThreads = 16;
	unsigned int *d_lfgReg;
	int *d_seq;
	cudaMalloc( (void**) &d_lfgReg, totalThreads *  LFG_REG_SIZE *4);
	cudaMalloc( (void**) &d_seq, totalThreads * 100 * 4);
	unsigned int *h_seq = (unsigned int *) malloc(totalThreads * 100 * 4);
	LFG175gmInit<<<4,4>>>(12345678,d_lfgReg);
	LFG175test<<<4,4>>>(d_lfgReg,d_seq);
	cudaMemcpy( h_seq, d_seq, totalThreads * 100 *4, cudaMemcpyDeviceToHost);

	for(int it=0;it<16;it++) {
		InitLfg175(12345678 + it);
		for(int i=0; i<100; i++) {
			local[i] = Lfg175();
		}
		printf("-------- thread id = %d --------\n",it);
		for(int i=0; i<20; i++) {
			printf("h: %10u  d: %10u\n",local[i],h_seq[i+100*it]);
		}
	}
	cudaFree( d_lfgReg );
	cudaFree( d_seq );
	free( h_seq	 );
}
